// -*- Mode: C++; tab-width: 4 -*-
// [c++ only for the emacs mode, c-mode is screwing up]

// four-register machine

#include <stdlib.h>
#include <stdio.h>
#include "machine.h"

unsigned char hexmap[] = {
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 0, 0, 0, 0, 0, 0, 0, 10, 11, 12,
  13, 14, 15, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

#ifdef COUNT_INSNS
unsigned long insn_use_table[256];
#endif

//
// This VM makes heavy use gcc extensions:
//    lexical functions
//    inline functions
//    labels as values (for threading)
//

int
machine (heap_type heap, SCHEME * entry_closure)
{
  SCHEME        acc             = 0;
  SCHEME *      val             = NULL;
  SCHEME *      lenv            = NULL;
  SCHEME *      k               = NULL;
  register BYTECODE *    code   = NULL;
  register ulong         pc     = 0;
  SCHEME *      current_heap    = heap.one;
  SCHEME *      other_heap      = heap.two;
  SCHEME *      free            = heap.free;
  const ulong   heap_size       = heap.size;
  SCHEME *      limit           = current_heap + (heap.size - 1024);
  SCHEME *      code_object     = NULL;
  SCHEME *      code_string     = NULL;

#include "opcodes.h"
#include "primops.h"

  // unsigned long long counter = 0;
  unsigned long counter = 0;

  // lexical functions
  
  // --------------------------------------------------
  // unpack an integer from the bytecode stream
  // --------------------------------------------------

#if 1
  inline long
    unpack_integer ()
    {
      long result = 0;
      while ((code[pc] != '+') && (code[pc] != '-')) {
        result = (result << 4) | hexmap[code[pc]];
        pc++;
      }
      if (code[pc] == '-') {
        result = - result;
      }
      pc++;
      return result;
    }
#else
  inline long
	unpack_integer()
	{
	  long result = *((long*)(code+pc));
	  pc += 4;
	  return result;
	}
#endif

  // --------------------------------------------------
  // allocate a tuple on the heap
  // --------------------------------------------------

  inline SCHEME *
    allocate (unsigned char tag, unsigned int size)
    {
      SCHEME * orig = free;
      int i;
      // space check here
      free[0] = (SCHEME) TAG_VALUE (tag, size);
      for (i=0; i < size; i++) {
        free[i+1] = 0;
      }
      free += (size+1);
      return orig;
    }

  // --------------------------------------------------
  // lexical variable reference
  // --------------------------------------------------

  inline SCHEME
    lexref (long depth, long index)
    {
      SCHEME * rib = lenv;
      while (depth--) {
        rib = (SCHEME *) rib[1];
      }
      return (SCHEME) rib[index+2];
    }

  // --------------------------------------------------
  // lexical variable assignment
  // --------------------------------------------------

  inline void
    lexset (long depth, long index, SCHEME value)
    {
      SCHEME * rib = lenv;
      while (depth--) {
        rib = (SCHEME *)rib[1];
      }
      rib[index+2] = value;
    }
  
  // --------------------------------------------------
  // cheney copying garbage collector
  // --------------------------------------------------

  SCHEME
    gc_flip ()
    {
      SCHEME * scan;

      inline int sitting_duck (SCHEME * p) {
        return (p >= current_heap) && (p < (current_heap + heap_size));
      }

      inline int copied_duck (SCHEME * p) {
        return (p >= other_heap)   && (p < (other_heap + heap_size));
      }

      SCHEME *
        copy (SCHEME * p)
        {
          SCHEME * pp = (SCHEME *) *p;
          if (IMMEDIATE (pp)) {
            return pp;
          } else if (sitting_duck (pp)) {
            if (copied_duck ((SCHEME *) (*pp))) {
              // pp points into to_space, return the forwarding address
              return (SCHEME *) (*pp);
            } else {
              unsigned char tc = GET_TYPECODE (*pp);
              if (tc <= TC_LAST) {
                // p points at an object in from_space, copy it
                SCHEME * addr = free;
                ulong length = GET_TUPLE_LENGTH (*pp);
                ulong i;
                // copy tag, children
                for (i=0; i < length+1; i++) {
                  *free++ = *pp++;
                }
                // leave a forwarding address where the tag used to be
                *(SCHEME *)(*p) = (SCHEME) addr;
                return addr;
              } else {
                fprintf (stderr, "panic: Unknown Type Code in GC copy loop: *pp=0x%08x tc=0x%x\n", (long)(*pp), tc);
				abort();
              }
            }
          } else {
            // pp points outside of the heap
            return pp;
          }
        }

	  fprintf (stderr, "[gc...");
      // place our roots
      scan = other_heap;
      free = scan + 6;
      scan[0] = (SCHEME) acc;
      scan[1] = (SCHEME) val;
      scan[2] = (SCHEME) lenv;
      scan[3] = (SCHEME) k;
      scan[4] = (SCHEME) code_object;
      scan[5] = (SCHEME) code_string;
      
      // unpad limit
      limit = current_heap + heap_size;


      // zero the to-space
      //      memset (free, 0, heap_size - (6*4));

      // copy the roots
      scan[0] = (SCHEME) copy (&(scan[0]));
      scan[1] = (SCHEME) copy (&(scan[1]));
      scan[2] = (SCHEME) copy (&(scan[2]));
      scan[3] = (SCHEME) copy (&(scan[3]));
      scan[4] = (SCHEME) copy (&(scan[4]));
      scan[5] = (SCHEME) copy (&(scan[5]));
      
      // bump scan
      scan += 6;

      // scan loop
      
      while (scan < free) {
        if (IMMEDIATE (*scan)) {
          scan++;
        } else {
          SCHEME * p = scan + 1;
          unsigned char tc = GET_TYPECODE (*scan);

          switch (tc) {

		  case TC_SAVE_VAL: // save_val = { tag, val, k }
          case TC_TUPLE:    // tuple    = { tag, next, arg0, arg1 ...}
          case TC_VECTOR:   // vector   = { tag, v0, v1, ... }
          case TC_CODE:     // code     = { tag, constant-vector, code-string }
          case TC_SYMBOL:   // symbol   = { tag, string }
          case TC_CLOSURE:  // closure  = { tag, code, pc, lenv }
          case TC_SAVE:     // save     = { tag, k, val, lenv, code, pc }
          case TC_PAIR: {   // pair     = { tag, car, cdr }
            // copy everything
            ulong i;
            ulong length = GET_TUPLE_LENGTH (*scan);
            for (i=0; i < length; i++) {
              // *p++ = (SCHEME) copy (p);
              *p = (SCHEME) copy (p);
              p++;
            }
            scan += length + 1;
          }
          break;

          case TC_STRING: {
            SCHEME length = GET_TUPLE_LENGTH (*scan);
            scan += length + 1;
          }
          break;

          default:
            fprintf (stderr, "panic: Unknown Type Code in GC scan loop: scan=0x%08x tc=0x%02x\n", (ulong)scan, tc);
			abort();
            break;
          }
        }
      }
      // swap heaps
      { SCHEME * temp = current_heap; current_heap = other_heap; other_heap = temp; }
      // XXX we should swap them in the heap structure, too?

      // replace roots
      acc           =            current_heap[0];
      val           = (SCHEME *) current_heap[1];
      lenv          = (SCHEME *) current_heap[2];
      k             = (SCHEME *) current_heap[3];
      code_object   = (SCHEME *) current_heap[4];
      code_string   = (SCHEME *) current_heap[5];

      // maintain invariant
      code = (BYTECODE *) (&(code_string[2]));

      // set new limit
      limit = current_heap + (heap_size - 1024);

      // zero the from-space
      memset (other_heap, 0, heap_size);

      fprintf (stderr, "collected %d words]\n", free - current_heap);
      // fprintf (stderr, "done]\n");
      return (SCHEME) BOX_INTEGER (free - current_heap);
    }

  // initialization
  code_object = (SCHEME *) entry_closure[1]; // closure = { tag, code, pc, lenv }
  code_string = (SCHEME *) code_object[2]; // code = { tag, constant-vector, code-string }
  code = (BYTECODE *)(&(code_string[2]));
  pc = UNBOX_INTEGER (entry_closure[2]);
  lenv = (SCHEME *) entry_closure[3];

  //--------------------------------------------------
  // Virtual Machine
  //--------------------------------------------------

#ifdef COUNT_INSNS
#define NEXT() { insn_use_table[code[pc]]++; counter++; goto *opcode_table[code[pc++]]; }
#else
# ifndef NO_COUNTER
#   define NEXT() counter++; goto *opcode_table[code[pc++]]
# else
#   define NEXT() goto *opcode_table[code[pc++]]
# endif
#endif 

  NEXT();

 TAG_OP_LIT: {
    acc = (SCHEME) unpack_integer();
  }
  NEXT();

 TAG_OP_LEXREF_00: acc = lexref (0,0); NEXT();
 TAG_OP_LEXREF_01: acc = lexref (0,1); NEXT();
 TAG_OP_LEXREF_02: acc = lexref (0,2); NEXT();
 TAG_OP_LEXREF_03: acc = lexref (0,3); NEXT();
 TAG_OP_LEXREF_10: acc = lexref (1,0); NEXT();
 TAG_OP_LEXREF_11: acc = lexref (1,1); NEXT();
 TAG_OP_LEXREF_12: acc = lexref (1,2); NEXT();
 TAG_OP_LEXREF_13: acc = lexref (1,3); NEXT();
 TAG_OP_LEXREF_20: acc = lexref (2,0); NEXT();
 TAG_OP_LEXREF_21: acc = lexref (2,1); NEXT();
 TAG_OP_LEXREF_22: acc = lexref (2,2); NEXT();
 TAG_OP_LEXREF_23: acc = lexref (2,3); NEXT();
 TAG_OP_LEXREF_30: acc = lexref (3,0); NEXT();
 TAG_OP_LEXREF_31: acc = lexref (3,1); NEXT();
 TAG_OP_LEXREF_32: acc = lexref (3,2); NEXT();
 TAG_OP_LEXREF_33: acc = lexref (3,3); NEXT();
 
 TAG_OP_VARREF: {
   long depth = unpack_integer();
   long index = unpack_integer();
   acc = lexref (depth, index);
 }
 NEXT();
 
 TAG_OP_ASSIGN: {
   long depth = unpack_integer();
   long index = unpack_integer();
   lexset (depth, index, acc);
 }
 NEXT();
 
 TAG_OP_CONSTANT: {
   long index = unpack_integer();
   // code = { tag, constant-vector, code-string }
   SCHEME * constant_vector = (SCHEME *) code_object[1];
   acc = constant_vector[index+1];
 }
 NEXT();
 
 TAG_OP_TEST: {
   long delta = unpack_integer();
   if ((unsigned long) acc == LJ_FALSE) {
     pc += delta;
   }
 }
 NEXT();
 
 TAG_OP_SAVE: {
   SCHEME * save = allocate (TC_SAVE, 5);
   int delta = unpack_integer();
   save[1]   = (SCHEME) k;
   save[2]   = (SCHEME) val;
   save[3]   = (SCHEME) lenv;
   save[4]   = (SCHEME) code_object;
   save[5]   = (SCHEME) BOX_INTEGER (pc + delta);
   k = save;
 }
 NEXT();
 
 TAG_OP_RESTORE: {
   if (k) {
	 if (GET_TYPECODE(k[0]) == TC_SAVE) {
     val		 = (SCHEME *) k[2];
     lenv        = (SCHEME *) k[3];
     code_object = (SCHEME *) k[4];
     code_string = (SCHEME *) code_object[2];
     code        = (BYTECODE *) (&(code_string[2]));
     pc          = UNBOX_INTEGER (k[5]);
     k           = (SCHEME *) k[1];
	 } else {
	   val = (SCHEME *) k[2];
	   k   = (SCHEME *) k[1];
	 }
   } else {
     fprintf (stderr, "\n%ld insns\n", counter);
     return (int) acc;
   }
 }
 NEXT();
 
 // experimenting with a cheaper save/restore for use by primops
 TAG_OP_SAVE_VAL: {
   SCHEME * save = allocate (TC_SAVE_VAL, 2);
   save[1]   = (SCHEME) k;
   save[2]   = (SCHEME) val;
   k = save;
 }
 NEXT();
 
 TAG_OP_BUILTIN: {
   ulong index = unpack_integer();
   acc = (SCHEME) ((*builtin_table[index])(val));
   if (!acc) {
     fprintf (stderr, "panic: error in builtin[%d]\n", index);
	 abort();
   }
 }
 NEXT();
 
 // allocate an extra word to hold env pointer
 TAG_OP_TUPLE_0: val = allocate (TC_TUPLE, 1); NEXT();
 TAG_OP_TUPLE_1: val = allocate (TC_TUPLE, 2); NEXT();
 TAG_OP_TUPLE_2: val = allocate (TC_TUPLE, 3); NEXT();
 TAG_OP_TUPLE_3: val = allocate (TC_TUPLE, 4); NEXT();
 TAG_OP_TUPLE: val = allocate (TC_TUPLE, 1 + unpack_integer()); NEXT();

 TAG_OP_STORE: {
   ulong index = unpack_integer();
   val[index] = acc;
 }
 NEXT();
 
 TAG_OP_STORE0: val[2]  = acc; NEXT();
 TAG_OP_STORE1: val[3]  = acc; NEXT();
 TAG_OP_STORE2: val[4]  = acc; NEXT();
 TAG_OP_STORE3: val[5]  = acc; NEXT();
 TAG_OP_STORE4: val[6]  = acc; NEXT();
 TAG_OP_STORE5: val[7]  = acc; NEXT();
 TAG_OP_STORE6: val[8]  = acc; NEXT();
 TAG_OP_STORE7: val[9]  = acc; NEXT();
 TAG_OP_STORE8: val[10] = acc; NEXT();
 TAG_OP_STORE9: val[11] = acc; NEXT();
 
 TAG_OP_JUMP:
 pc = pc + unpack_integer();
 NEXT();
 
 TAG_OP_PUSH_ENV:
 val[1] = (SCHEME) lenv;
 lenv = val;
 NEXT();
 
 TAG_OP_PROC: {
   SCHEME * closure = allocate (TC_CLOSURE, 3);
   int delta = unpack_integer();
   closure[1] = (SCHEME) code_object;
   closure[2] = (SCHEME) BOX_INTEGER (pc);
   closure[3] = (SCHEME) lenv;
   acc = (SCHEME) closure;
   pc += delta;
 }
 NEXT();
 
 TAG_OP_INVOKE: {
   SCHEME * closure = (SCHEME *)acc;
   if ((!IMMEDIATE(closure)) && (GET_TYPECODE(*closure) == TC_CLOSURE)) {
     code_object = (SCHEME *) closure[1]; // closure = {tag, code, pc, lenv}
     code_string = (SCHEME *) code_object[2];
     code = (BYTECODE *) (&(code_string[2]));
     pc = UNBOX_INTEGER (closure[2]);
     val[1] = (SCHEME) closure[3];       // extend environment
     lenv = val;
     // check heap
     if (free > limit) {
       gc_flip();
     }
   } else {
     fprintf (stderr, "panic: ");
     builtin_print_object (acc);
     fprintf (stderr, ": not a closure\n");
	 abort();
   }
 }
 NEXT();
 
 TAG_OP_PRIMOP: {
   enum primops primop = code[pc++];
   // val = {tag, next, arg0, arg1, ....}
   long *arg = (long *)(&val[2]);
   goto *primop_table[primop];
   
 TAG_PRIM_ADD:
   // 2(x+y)+1 == 2x+2y + 1 == (2x+1) + (2y+1) - 1
   // acc = (arg[0] + arg[1]) - 1;
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) + UNBOX_INTEGER (arg[1]));
   NEXT();
 TAG_PRIM_SUB:
   // 2(x-y)+1 == 2x-2y + 1 == (2x+1) - (2y+1) + 1
   // acc = (arg[0] - arg[1]) + 1;
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) - UNBOX_INTEGER (arg[1]));
   NEXT();
 TAG_PRIM_MUL:
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) * UNBOX_INTEGER(arg[1]));
   NEXT();
 TAG_PRIM_DIV:
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) / UNBOX_INTEGER(arg[1]));
   NEXT();
 TAG_PRIM_MOD: 
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) % UNBOX_INTEGER(arg[1]));
   NEXT();
 TAG_PRIM_OR:
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) | UNBOX_INTEGER(arg[1]));
   NEXT();
 TAG_PRIM_AND:
   acc = BOX_INTEGER (UNBOX_INTEGER(arg[0]) && UNBOX_INTEGER(arg[1]));
   NEXT();
 TAG_PRIM_EQ:
   acc = ((arg[0] == arg[1]) ? LJ_TRUE : LJ_FALSE);
   NEXT();
 TAG_PRIM_GT:
   acc = ((arg[0] > arg[1])  ? LJ_TRUE : LJ_FALSE);
   NEXT();
 TAG_PRIM_LT:
   acc = ((arg[0] < arg[1])  ? LJ_TRUE : LJ_FALSE);
   NEXT();
 TAG_PRIM_GE:
   acc = ((arg[0] >= arg[1]) ? LJ_TRUE : LJ_FALSE);
   NEXT();
 TAG_PRIM_LE:
   acc = ((arg[0] <= arg[1]) ? LJ_TRUE : LJ_FALSE);
   NEXT();
 TAG_PRIM_LEFT_SHIFT:
   acc = BOX_INTEGER ((UNBOX_INTEGER(arg[0]))<<(UNBOX_INTEGER(arg[1])));
   NEXT();
 TAG_PRIM_RIGHT_SHIFT:
   acc = BOX_INTEGER ((UNBOX_INTEGER(arg[0]))>>(UNBOX_INTEGER(arg[1])));
   NEXT();
   
 TAG_PRIM_CONS: {
     SCHEME * pair = allocate (TC_PAIR, 2);
     pair[1] = (SCHEME)arg[0];
     pair[2] = (SCHEME)arg[1];
     acc = (SCHEME) pair;
   }
   NEXT();
   
 TAG_PRIM_CAR: {
     SCHEME * pair = (SCHEME *) arg[0];
     if (!IS_TYPE(TC_PAIR, *pair)) {
       fprintf (stderr, "warning: (car <non-pair>)");
       builtin_print_object (pair);
       fflush (stdout);
     }
     acc = ((SCHEME *)pair)[1];
   }
   NEXT();
   
 TAG_PRIM_CDR: {
     SCHEME * pair = (SCHEME *) arg[0];
     if (!IS_TYPE(TC_PAIR, *pair)) {
       fprintf (stderr, "warning: (cdr <non-pair>)");
       builtin_print_object (pair);
       fflush (stdout);
     }
     acc = ((SCHEME *)pair)[2];
   }
   NEXT();
   
 TAG_PRIM_APPLY1: {
     // ick, a lot of code shared with OP_INVOKE
     SCHEME * proc = (SCHEME *) arg[0];
     SCHEME * args = (SCHEME *) arg[1];
     if ((!IMMEDIATE(proc)) && (GET_TYPECODE(*proc) == TC_CLOSURE)) {
       code_object = (SCHEME *) proc[1];
       code_string = (SCHEME *) code_object[2];
       code = (BYTECODE *) (&(code_string[2]));
       pc = UNBOX_INTEGER (proc[2]);
       args[1] = (SCHEME) proc[3]; // extend environment
       lenv = args;
       // check heap
       if (free > limit) {
         gc_flip();
       }
     } else {
       fprintf (stderr, "panic: ");
       builtin_print_object (acc);
       fprintf (stderr, ": not a closure\n");
	   abort();
     }
   }
   NEXT();

 TAG_PRIM_TUPLE_REF: {
     SCHEME * tuple = (SCHEME *) arg[0];
     ulong index = (ulong) UNBOX_INTEGER (arg[1]);
     acc = tuple[1+index];
   }
   NEXT();

 TAG_PRIM_TUPLE_SET: {
     SCHEME * tuple = (SCHEME *) arg[0];
     ulong len = GET_TUPLE_LENGTH (*tuple);
     ulong index = (ulong) UNBOX_INTEGER (arg[1]);
     SCHEME value = arg[2];
     if (index < len) {
       tuple[1+index] = value;
       acc = LJ_UNDEFINED;
     } else {
       fprintf (stderr, "panic: tuple-set! out of range\n");
	   abort();
     }
   }
   NEXT();

 TAG_PRIM_NEW_TUPLE: {
     SCHEME * tuple = allocate (UNBOX_INTEGER (arg[0]), UNBOX_INTEGER (arg[1]));
     acc = (SCHEME) tuple;
   }
   NEXT();

 TAG_PRIM_BUILD_TUPLE: {
     // a generalized 'cons'
     // (%build-tuple <tc> <elem1> <elem2> ...)
     ulong len = GET_TUPLE_LENGTH (*val) - 2;
     SCHEME * tuple = allocate (UNBOX_INTEGER (arg[0]), len);
     while (len) {
       tuple[len] = arg[len];
       len--;
     }
     acc = (SCHEME) tuple;
   }
   NEXT();

 TAG_PRIM_TUPLE_LENGTH: {
     // XXX typecheck
     SCHEME * tuple = (SCHEME *)arg[0];
     acc = BOX_INTEGER (GET_TUPLE_LENGTH(*tuple));
   }
   NEXT();

 TAG_PRIM_GET_TYPECODE: {
     SCHEME * x = (SCHEME *) arg[0];
     if (IMMEDIATE(x)) {
       if (IS_INTEGER(x)) {
         acc = BOX_INTEGER(0);
       } else {
         acc = BOX_INTEGER (GET_TYPECODE (x));
       }
     } else {
       acc = BOX_INTEGER (GET_TYPECODE (*x));
     }
   }
   NEXT();

   // XXX typecheck
 TAG_PRIM_INTEGER_CHAR:
   acc = TAG_VALUE (TC_CHAR, UNBOX_INTEGER (arg[0]));
   NEXT();

   // XXX typecheck
 TAG_PRIM_CHAR_INTEGER:
   acc = BOX_INTEGER (GET_PAYLOAD (arg[0]));
   NEXT();

 TAG_PRIM_GET_ENVIRONMENT:
   acc = (SCHEME) lenv;
   NEXT();
   
   //--------------------------------------------------
   // call/cc
   //--------------------------------------------------
   
 TAG_PRIM_GETCC:
   acc = (SCHEME) k;
   NEXT();
   
 TAG_PRIM_PUTCC:
   k = (SCHEME *) arg[0]; k = (SCHEME *) k[1]; acc = arg[1];
   NEXT();
   
   //--------------------------------------------------
   // heap-related primops
   //--------------------------------------------------
   
 TAG_PRIM_GC:
   acc = gc_flip();
   NEXT();

 TAG_PRIM_ROOM:
   acc = BOX_INTEGER ((char *) limit - (char *) free);
   NEXT();

 TAG_PRIM_INSNS:
   acc = BOX_INTEGER (counter);
   NEXT();

 TAG_PRIM_DUMP_IMAGE: {
     // dump an image
     SCHEME * thunk;
     SCHEME * from_space;

     FILE * dump_file = fopen ("dump.image", "wb");
     fprintf (dump_file, "(lumberjack-image (bytecode))\n");

     // do a gc; for a compact dump
     gc_flip();
     from_space = current_heap;
     // skip the six roots
     from_space = from_space + 5;
     thunk = (SCHEME *) val[2]; // gc has moved val
     from_space[0] = (SCHEME) thunk;

     // relocate the heap to a zero offset (XXX is this necessary?, why can't
     // we just dump it along with the base address??)
     gc_relocate (1, from_space, (SCHEME *) free, - (size_t) from_space);
     // dump it
     fwrite ((const void *) from_space, sizeof(SCHEME), (size_t) (free - from_space), dump_file);
     // write offset to thunk
     fwrite (&(from_space[0]), sizeof(SCHEME), 1, dump_file);
     fclose (dump_file);
     // unrelocate heap
     gc_relocate (1, from_space, (SCHEME *) free, (size_t) from_space);
     acc = LJ_TRUE;
   }
   NEXT();

 TAG_PRIM_UNDEFINED:
   acc = LJ_UNDEFINED;
   NEXT();

 TAG_PRIM_UNKNOWN: {
     fprintf (stderr, "panic: unsupported primop '%c'\n", primop);
	 abort();
     NEXT();
   }
   NEXT();
 }

 TAG_OP_UNKNOWN: {
   enum opcodes opcode = code[pc-1];
   fprintf (stderr, "panic: unknown opcode: %c/%d\n", opcode, opcode);
   abort();
   NEXT();
 }
}

//
// relocate a heap's pointers in-place.  the operation is reversible
// by calling gc_relocate again with delta = -delta
//

SCHEME
gc_relocate (SCHEME nroots, SCHEME * start, SCHEME * finish, size_t delta)
{
  SCHEME * scan = start;
  int i;

#define ADJUST(p)                               \
  do {                                          \
    if ((*p) && (!IMMEDIATE(*(p)))) {           \
      *(p) += delta;                            \
    }                                           \
  } while (0)

  // roots are either immediate values, or pointers to tuples
  // (map ADJUST roots)

  for (i=0; i < nroots; i++, scan++) {
    ADJUST(scan);
  }

  while (scan < finish) {
    // There must be a tuple here
    int tc = GET_TYPECODE (*scan);
    SCHEME * p = scan + 1;

    // printf ("scan=%08x, *scan=%08x\n", scan, *scan);

    switch (tc) {

    case TC_TUPLE:  // tuple   = { tag, next, arg0, arg1 ...}
    case TC_VECTOR: // vector  = { tag, v0, v1, ... }
    case TC_CODE:   // code    = { tag, constant-vector, code-string }
    case TC_SYMBOL: // symbol  = { tag, string }
    case TC_CLOSURE:// closure = { tag, code, pc, lenv }
    case TC_SAVE:   // save    = { tag, k, val, lenv, code, pc }
    case TC_PAIR: { // pair    = { tag, car, cdr }
      // adjust everything
      int i;
      SCHEME length = GET_TUPLE_LENGTH (*scan);
      for (i=0; i < length; i++, p++) {
        ADJUST (p);
      }
      scan += length+1;
    }
    break;

    case TC_STRING: {
      SCHEME length = GET_TUPLE_LENGTH (*scan);
      scan += length+1;
    }
    break;

    default:
      printf ("Unknown Type Code in heap-relocation loop: scan=0x%08x tc=0x%x\n", (SCHEME)scan, tc);
	  abort();
      break;
    }
  }
}

SCHEME
run_image (char * image_file, const size_t heap_size)
{
  heap_type heap;
  FILE * image;
  SCHEME * entry_thunk = NULL;
  ulong entry_thunk_offset;
  ulong header_offset;
  ulong image_size;
  
  // allocate the heap
  heap.one = (SCHEME *) calloc (sizeof(SCHEME), heap_size);
  heap.two = (SCHEME *) calloc (sizeof(SCHEME), heap_size);
  heap.size = heap_size;

  // load the image file
  image = fopen (image_file, "rb");

  if (!image) {
    fprintf (stderr, "failed to open file %s\n", image_file);
    exit (-1);
  }

  // seek to the end, where our entry-thunk offset awaits us.
  fseek (image, - sizeof (SCHEME), SEEK_END);
  // compute the size of the image proper
  image_size = ftell (image);

  // read the offset to the entry thunk (a closure)
  fread (&entry_thunk_offset, sizeof(SCHEME), 1, image);

  // back to the start of the file
  fseek (image, 0, SEEK_SET);
  // skip over header
  read_header (image);
  header_offset = ftell (image);
  image_size -= header_offset;

  fprintf (stderr, "read %d bytes\n", fread (heap.one, 1, image_size, image));
  
  // use entry thunk as root for gc/relocator
  heap.one[0] = entry_thunk_offset;
  
  // relocate the image
  gc_relocate (1, heap.one, (SCHEME*)((unsigned char*)heap.one + image_size), (SCHEME) heap.one);

  // set the free pointer
  heap.free = (SCHEME *) ((unsigned char *)heap.one + image_size);

  entry_thunk = heap.one + (entry_thunk_offset/sizeof(SCHEME));
  return machine (heap, entry_thunk);
}

int
read_header (FILE * file)
{
  int depth = 0;
  // tiny lisp 'skipper' (as opposed to 'reader')
  do {
    char ch = fgetc(file);
    switch (ch) {
    case '(':
      depth++;
      break;
    case ')':
      depth--;
      break;
    case '"':
      while (fgetc(file) != '"') {
        // empty body
      }
      break;
    default:
      break;
    }
  } while (depth);
  // read terminating newline
  fgetc(file);
}

int
main (int argc, char * argv[])
{
  FILE * code_file = NULL;
  if (argc > 1) {
    run_image (argv[1], 4 * 1024 * 1024);
#ifdef COUNT_INSNS
	{
	  int i;
	  for (i=0; i<256; i++) {
		if (insn_use_table[i] > 0) {
		  fprintf (stderr, "%d %d\n",  insn_use_table[i], i, i);
		}
	  }
	}
#endif
  } else {
    fprintf (stderr, "%s <image-file>\n", argv[0]);
  }
}
