;; -*- Mode: Scheme; tab-width: 4 -*-

(define (lex-ref env addr)
  (let ((depth (car addr))
	(index (cadr addr)))
    (vector-ref (list-ref env depth) index)))

(define (lex-set! env addr val)
  (let ((depth (car addr))
	(index (cadr addr)))
    (vector-set! (list-ref env depth) index val)))

(define (machine code)
  (let ((acc #f)
		(val* #f)
		(lenv '())
		(k '())
		(pc 0))
    (call-with-current-continuation
     (lambda (halt)
       (let loop ()
		 (let ((insn (vector-ref code pc)))
		   (display pc) (display " --- ") (display insn) (newline)
		   (case (car insn)
			 ((literal) (set! acc (second insn)))
			 ((new-tuple) (set! val* (make-vector (second insn))))
			 ((tuple-store) (vector-set! val* (second insn) acc))
			 ((varref) (set! acc (lex-ref lenv (second insn))))
			 ((assign) (lex-set! lenv (second insn) acc))
			 ((primop)
			  (case (second insn)
				((%+) (set! acc (apply + (vector->list val*))))
				((%-) (set! acc (apply - (vector->list val*))))
				((%*) (set! acc (apply * (vector->list val*))))
				((%/) (set! acc (apply / (vector->list val*))))
				((%eq?) (set! acc (eq? (vector-ref val* 0) (vector-ref val* 1))))
				((%ge?) (set! acc (>= (vector-ref val* 0) (vector-ref val* 1))))
				((%cons) (set! acc (cons (vector-ref val* 0) (vector-ref val* 1))))
				((%print) (display val*))
				(else (error "Unknown Primop"))))
			 ((test)
			  (if (not acc)
				  (set! pc (+ 1 (+ pc (second insn))))))
			 ((jump)
			  (set! pc (+ pc (second insn))))
			 ((close)
			  (set! acc (list code pc lenv))
			  (set! pc (+ pc (second insn))))
			 ((invoke)
			  (set! code (first acc))
			  (set! pc (second acc))
			  (set! lenv (cons val* (third acc))))
			 ((save)
			  (set! k (cons (list val* lenv code (+ pc (second insn))) k)))
			 ((restore)
			  (if (null? k)
				  (halt)
				  (let ((saved (car k)))
					(set! val* (first saved))
					(set! lenv (second saved))
					(set! code (third saved))
					(set! pc (fourth saved))
					(set! k (cdr k)))))
			 ((push-env)
			  (set! lenv (cons val* lenv)))
			 ((new-tuple)
			  (set! val* (make-vector (second insn))))
			 ((tuple-store)
			  (vector-set! val* (second insn) acc))
			 (else
			  (error "Unknown opcode"))
			 )
		   (display "acc ") (display acc) (newline)
		   (display "val* ") (display val*) (newline)
		   (display "lenv ") (display lenv) (newline)
		   (display "k ") (if (not (null? k))
							  (begin (display (caar k)) (display " ") (display (cadar k)) (newline))
							  (newline))
		   )
		 (set! pc (+ pc 1))
		 (loop))))
    acc))

;; [taken out of compiler.scm]

  ;; ===========================================================================
  ;; Vectorized Code
  ;; ===========================================================================
  
  ;; The output from vectorize can be fed to the Scheme version of the VM.
;  (define (vectorize code)
;    (let loop ((code code))
;      (cond ((null? code) '())
;            (else
;             (let ((insn (car code)))
;               (case (car insn)
;                 ((test)
;                  (let ((then-code (vectorize (second insn)))
;                        (else-code (vectorize (third insn))))
;                    (append
;                     `((test ,(length then-code))) then-code
;                     `((jump ,(length else-code))) else-code
;                     (loop (cdr code)))))
;                 ((close)
;                  (let ((proc-code (vectorize (second insn))))
;                    (append (cons `(close ,(length proc-code)) proc-code) (loop (cdr code)))))
;                 ((save)
;                  (let ((proc-code (vectorize (second insn))))
;                    (append (cons `(save ,(length proc-code)) proc-code) (loop (cdr code)))))
;                 (else
;                  (cons insn (loop (cdr code)))))
;               )))))
  
;(define (compile-svm exp)
;  (list->vector (vectorize (compile exp))))
